@@ -1,5 +1,5 @@
 package Thrift::API::HiveClient2;
-$Thrift::API::HiveClient2::VERSION = '0.017';
+$Thrift::API::HiveClient2::VERSION = '0.019';
 {
   $Thrift::API::HiveClient2::DIST = 'Thrift-API-HiveClient2';
 }
@@ -8,10 +8,10 @@ $Thrift::API::HiveClient2::VERSION = '0.017';
 
 use strict;
 use warnings;
-
 use Moo;
 use Carp;
 use Scalar::Util qw( reftype blessed );
+use List::MoreUtils 'zip';
 
 use Thrift;
 use Thrift::Socket;
@@ -21,6 +21,36 @@ use Thrift::BufferedTransport;
 
 use Thrift::API::HiveClient2::TCLIService;
 
+# See https://msdn.microsoft.com/en-us/library/ms711683(v=vs.85).aspx
+my @odbc_coldesc_fields = qw(
+    TABLE_CAT
+    TABLE_SCHEM
+    TABLE_NAME
+    COLUMN_NAME
+    DATA_TYPE
+    TYPE_NAME
+    COLUMN_SIZE
+    BUFFER_LENGTH
+    DECIMAL_DIGITS
+    NUM_PREC_RADIX
+    NULLABLE
+    REMARKS
+    COLUMN_DEF
+    SQL_DATA_TYPE
+    SQL_DATETIME_SUB
+    CHAR_OCTET_LENGTH
+    ORDINAL_POSITION
+    IS_NULLABLE
+);
+
+my @tabledesc_fields = qw(
+    TABLE_CAT
+    TABLE_SCHEM
+    TABLE_NAME
+    TABLE_TYPE
+    REMARKS
+    );
+
 # Don't use XS for now, fails initializing properly with BufferedTransport. See
 # Thrift::XS documentation.
 has use_xs => (
@@ -187,11 +217,18 @@ sub _build_session_handle {
 }
 
 has _operation => (
-    is => "rwp",
-    isa     => sub {
-        die "Operation isn't a Thrift::API::HiveClient2::TExecuteStatementResp"
-            if defined $_[0] && ( ! blessed($_[0]) || !$_[0]->isa('Thrift::API::HiveClient2::TExecuteStatementResp') ) },
-    lazy    => 1,
+    is  => "rwp",
+    isa => sub {
+        die "Operation isn't a Thrift::API::HiveClient2::T*Resp"
+            if defined $_[0]
+            && (
+            !blessed( $_[0] )
+            || (   !$_[0]->isa('Thrift::API::HiveClient2::TExecuteStatementResp')
+                && !$_[0]->isa('Thrift::API::HiveClient2::TGetColumnsResp')
+                && !$_[0]->isa('Thrift::API::HiveClient2::TGetTablesResp') )
+            );
+    },
+    lazy => 1,
 );
 
 has _operation_handle => (
@@ -243,11 +280,17 @@ sub execute {
 {
     # cache the column names we need to extract from the bloated data structure
     # (keyed on query)
-    my $column_keys;
+    my ($column_keys, $column_names);
+
+    sub fetch_hashref {
+        my $self = shift;
+        my ( $rv, $rows_at_a_time ) = @_;
+        return $self->fetch($rv, $rows_at_a_time, 1);
+    }
 
     sub fetch {
         my $self = shift;
-        my ( $rv, $rows_at_a_time ) = @_;
+        my ( $rv, $rows_at_a_time, $use_hashref ) = @_;
 
         # if $rv looks like a number, use it instead of $rows_at_a_time
         $rows_at_a_time = $rv if !$rows_at_a_time && $rv =~ /^[1-9][0-9]*$/;
@@ -275,6 +318,20 @@ sub execute {
 
                 # Find which fields to extract from each row, only on the first iteration
                 if ( !@{ $column_keys->{ $rv } || [] } ) {
+
+                    # metadata for the query
+                    if ($use_hashref) {
+                        my $rh_meta = $self->_client->GetResultSetMetadata(
+                            Thrift::API::HiveClient2::TGetResultSetMetadataReq->new({
+                                operationHandle => $self->_operation_handle
+                            })
+                        );
+                        $column_names =
+                            [ map { $_->{columnName} } @{ $rh_meta->{schema}{columns} || [] } ];
+                    }
+
+                    # TODO redo all this using the TGetResultSetMetadataResp object we retrieved
+                    # above
                     for my $column ( @{ $row->{colVals} || [] } ) {
 
                         my $first_col = {%$column};
@@ -291,22 +348,104 @@ sub execute {
                     }
                 }
 
-                # TODO find something faster?
+                # TODO find something faster? (see comment above)
 
                 my $idx = 0;
-                push @$result,
-                    [
-                        map  { $_->value  }
+                my $retval = [
+                    map { $_->value }
                         grep { defined $_ }
-                        map  { $row->{colVals}[ $idx++ ]{$_} }
-                        @{ $column_keys->{$rv} }
-                    ];
+                            map  { $row->{colVals}[ $idx++ ]{$_} }
+                                @{ $column_keys->{$rv} }
+                ];
+                if ($use_hashref) {
+                    push @$result, { zip @$column_names, @$retval };
+                }
+                else {
+                    push @$result, $retval;
+                }
             }
         }
         return wantarray ? ( $result, $has_more_rows ) : ( @$result ? $result : undef );
     }
 }
 
+sub get_columns {
+    my $self = shift;
+    my ( $table, $schema ) = @_;
+
+    # note that not specifying a table name would return all columns for all
+    # tables we probably don't want that, but feel free to change this
+    # behaviour. Same goes for the schema name: we probably want a default
+    # value for the schema, which is what we use here.
+    die "Unspecified table name" if !$table;
+    $schema //= "default";
+
+    $self->_cleanup_previous_operation;
+
+    my $rh = $self->_client->GetColumns(
+        Thrift::API::HiveClient2::TGetColumnsReq->new(
+            {   sessionHandle => $self->_session_handle,
+                catalogName   => undef,
+                schemaName    => $schema,
+                tableName     => $table,
+                columnName    => undef,
+                confOverlay   => {}
+            }
+        )
+    );
+    if ( $rh->{status}{errorCode} ) {
+        die __PACKAGE__ . "::execute: $rh->{status}{errorMessage}";
+    }
+    $self->_set__operation($rh);
+    $self->_set__operation_handle( $rh->{operationHandle} );
+    my $columns;
+    while ( my $res = $self->fetch($rh) ) {
+        for my $line (@$res) {
+            my $idx = 0;
+            push @$columns, { map { $_ => $line->[ $idx++ ] } @odbc_coldesc_fields };
+        }
+    }
+    return $columns;
+}
+
+sub get_tables {
+    my $self = shift;
+    my ( $schema, $table_pattern ) = @_;
+
+    # note that not specifying a table name would return all columns for all
+    # tables we probably don't want that, but feel free to change this
+    # behaviour. Same goes for the schema name: we probably want a default
+    # value for the schema, which is what we use here.
+    $schema //= "default";
+
+    $self->_cleanup_previous_operation;
+
+    my $rh = $self->_client->GetTables(
+        Thrift::API::HiveClient2::TGetTablesReq->new(
+            {
+                sessionHandle => $self->_session_handle,
+                catalogName   => undef,
+                schemaName    => $schema,
+                tableName     => $table_pattern,
+                confOverlay   => {},
+            }
+        )
+    );
+    if ( $rh->{status}{errorCode} ) {
+        die __PACKAGE__ . "::execute: $rh->{status}{errorMessage}";
+    }
+    $self->_set__operation($rh);
+    $self->_set__operation_handle( $rh->{operationHandle} );
+    my $tables;
+    while ( my $res = $self->fetch($rh) ) {
+        for my $line (@$res) {
+            my $idx = 0;
+            push @$tables, { map { $_ => $line->[ $idx++ ] } @tabledesc_fields };
+        }
+    }
+    return $tables;
+}
+
 sub DEMOLISH {
     my $self = shift;
 
@@ -353,7 +492,7 @@ Thrift::API::HiveClient2 - Perl to HiveServer2 Thrift API wrapper
 
 =head1 VERSION
 
-version 0.017
+version 0.019
 
 =head1 METHODS
 
@@ -411,6 +550,37 @@ number of rows to be retrieved if it looks like a positive integer:
 
      my $rv = $client->fetch( 10_000 );
 
+=head2 fetch_hashref
+
+Same use as above, but result is returned as an arrayref of hashes (which keys are
+the column names)
+
+=head2 get_columns
+
+Get the columns description for a table, returned in an array of hashrefs which keys are named after the result of an
+ODBC GetColumns call. "default" is used for the schema name is none is specified as 2nd argument. The hashes keys
+documentation can be found on https://msdn.microsoft.com/en-us/library/ms711683(v=vs.85).aspx for instance.
+
+    my $columns = $client->get_columns('<table name>'[, '<schema name>']);
+
+=head2 get_tables
+
+Get a list of tables. Optional table name pattern as a first argument (use undef or '%' to get all tables while
+defining a schema as a second argument), and optional schema second arg (default is "default")
+
+    my $tables = $client->get_tables(['<table pattern, SQL wildcards accepted>', ['<schema name>']]);
+
+Returns an arrayref of hashes:
+
+    [...
+    {
+        'REMARKS' => 'test comment', # table comment
+        'TABLE_NAME' => 'foo_bar',   # table name
+        'TABLE_SCHEM' => 'default',  # schema ("database")
+        'TABLE_TYPE' => 'TABLE',     # TABLE, VIEW, etc
+        'TABLE_CAT' => '',           # catalog (unused?)
+    }];
+
 =head1 WARNING
 
 Thrift in Perl originally did not support SASL, so authentication needed to be
@@ -463,7 +633,7 @@ David Morel <david.morel@amakuru.net>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2014 by David Morel & Booking.com. Portions are (c) R.Scaffidi, Thrift files are (c) Apache Software Foundation..
+This software is Copyright (c) 2015 by David Morel & Booking.com. Portions are (c) R.Scaffidi, Thrift files are (c) Apache Software Foundation..
 
 This is free software, licensed under:
 
